@@ -18,7 +18,7 @@ niceness=0
 logging=init
 want_syslog=0
 syslog_tag=
-user=@MYSQLD_USER@
+user='@MYSQLD_USER@'
 pid_file=
 err_log=
 
@@ -27,7 +27,28 @@ syslog_tag_mysqld_safe=mysqld_safe
 
 trap '' 1 2 3 15			# we shouldn't let anyone kill us
 
-umask 007
+# MySQL-specific environment variable. First off, it's not really a umask,
+# it's the desired mode. Second, it follows umask(2), not umask(3) in that
+# octal needs to be explicit. Our shell might be a proper sh without printf,
+# multiple-base arithmetic, and binary arithmetic, so this will get ugly.
+# We reject decimal values to keep things at least half-sane.
+umask 007                               # fallback
+UMASK="${UMASK-0640}"
+fmode=`echo "$UMASK" | sed -e 's/[^0246]//g'`
+octalp=`echo "$fmode"|cut -c1`
+fmlen=`echo "$fmode"|wc -c|sed -e 's/ //g'`
+if [ "x$octalp" != "x0" -o "x$UMASK" != "x$fmode" -o "x$fmlen" != "x5" ]
+then
+  fmode=0640
+  echo "UMASK must be a 3-digit mode with an additional leading 0 to indicate octal." >&2
+  echo "The first digit will be corrected to 6, the others may be 0, 2, 4, or 6." >&2
+fi
+fmode=`echo "$fmode"|cut -c3-4`
+fmode="6$fmode"
+if [ "x$UMASK" != "x0$fmode" ]
+then
+  echo "UMASK corrected from $UMASK to 0$fmode ..."
+fi
 
 defaults=
 case "$1" in
@@ -64,6 +85,7 @@ my_which ()
 {
   save_ifs="${IFS-UNSET}"
   IFS=:
+  ret=0
   for file
   do
     for dir in $PATH
@@ -74,15 +96,18 @@ my_which ()
         continue 2
       fi
     done
-    return 1  # Failure, didn't find file in path
+
+	ret=1  #signal an error
+	break
   done
+
   if [ "$save_ifs" = UNSET ]
   then
     unset IFS
   else
     IFS="$save_ifs"
   fi
-  return 0  # Success
+  return $ret  # Success
 }
 
 log_generic () {
@@ -179,6 +204,7 @@ parse_arguments() {
         ;;
       --nice=*) niceness="$val" ;;
       --open-files-limit=*) open_files="$val" ;;
+      --open_files_limit=*) open_files="$val" ;;
       --skip-kill-mysqld*) KILL_MYSQLD=0 ;;
       --syslog) want_syslog=1 ;;
       --skip-syslog) want_syslog=0 ;;
@@ -212,19 +238,30 @@ fi
 
 MY_PWD=`pwd`
 # Check for the directories we would expect from a binary release install
-if test -f "$relpkgdata"/english/errmsg.sys -a -x ./bin/mysqld
+if test -n "$MY_BASEDIR_VERSION" -a -d "$MY_BASEDIR_VERSION"
 then
-  MY_BASEDIR_VERSION=$MY_PWD		# Where bin, share and data are
-  ledir=$MY_BASEDIR_VERSION/bin		# Where mysqld is
+  # BASEDIR is already overridden on command line.  Do not re-set.
+
+  # Use BASEDIR to discover le.
+  if test -x "$MY_BASEDIR_VERSION/libexec/mysqld"
+  then
+    ledir="$MY_BASEDIR_VERSION/libexec"
+  else
+    ledir="$MY_BASEDIR_VERSION/bin"
+  fi
+elif test -f "$relpkgdata"/english/errmsg.sys -a -x "$MY_PWD/bin/mysqld"
+then
+  MY_BASEDIR_VERSION="$MY_PWD"		# Where bin, share and data are
+  ledir="$MY_PWD/bin"			# Where mysqld is
 # Check for the directories we would expect from a source install
-elif test -f "$relpkgdata"/english/errmsg.sys -a -x ./libexec/mysqld
+elif test -f "$relpkgdata"/english/errmsg.sys -a -x "$MY_PWD/libexec/mysqld"
 then
-  MY_BASEDIR_VERSION=$MY_PWD		# Where libexec, share and var are
-  ledir=$MY_BASEDIR_VERSION/libexec	# Where mysqld is
+  MY_BASEDIR_VERSION="$MY_PWD"		# Where libexec, share and var are
+  ledir="$MY_PWD/libexec"		# Where mysqld is
 # Since we didn't find anything, used the compiled-in defaults
 else
-  MY_BASEDIR_VERSION=@prefix@
-  ledir=@libexecdir@
+  MY_BASEDIR_VERSION='@prefix@'
+  ledir='@libexecdir@'
 fi
 
 
@@ -274,7 +311,10 @@ export MYSQL_HOME
 
 # Get first arguments from the my.cnf file, groups [mysqld] and [mysqld_safe]
 # and then merge with the command line arguments
-if test -x ./bin/my_print_defaults
+if test -x "$MY_BASEDIR_VERSION/bin/my_print_defaults"
+then
+  print_defaults="$MY_BASEDIR_VERSION/bin/my_print_defaults"
+elif test -x ./bin/my_print_defaults
 then
   print_defaults="./bin/my_print_defaults"
 elif test -x @bindir@/my_print_defaults
@@ -352,6 +392,12 @@ then
   # Log to err_log file
   log_notice "Logging to '$err_log'."
   logging=file
+
+  if [ ! -e "$err_log" ]; then                  # if error log already exists,
+    touch "$err_log"                            # we just append. otherwise,
+    chmod "$fmode" "$err_log"                   # fix the permissions here!
+  fi
+
 else
   if [ -n "$syslog_tag" ]
   then
@@ -373,16 +419,20 @@ then
   fi
   # Change the err log to the right user, if it is in use
   if [ $want_syslog -eq 0 ]; then
-    touch $err_log
-    chown $user $err_log
+    touch "$err_log"
+    chown $user "$err_log"
   fi
   if test -n "$open_files"
   then
     ulimit -n $open_files
-    append_arg_to_args "--open-files-limit=$open_files"
   fi
 fi
 
+if test -n "$open_files"
+then
+  append_arg_to_args "--open-files-limit=$open_files"
+fi
+
 safe_mysql_unix_port=${mysql_unix_port:-${MYSQL_UNIX_PORT:-@MYSQL_UNIX_ADDR@}}
 # Make sure that directory for $safe_mysql_unix_port exists
 mysql_unix_port_dir=`dirname $safe_mysql_unix_port`
@@ -399,7 +449,7 @@ then
   MYSQLD=mysqld
 fi
 
-if test ! -x $ledir/$MYSQLD
+if test ! -x "$ledir/$MYSQLD"
 then
   log_error "The file $ledir/$MYSQLD
 does not exist or is not executable. Please cd to the mysql installation
@@ -411,7 +461,7 @@ fi
 
 if test -z "$pid_file"
 then
-  pid_file=$DATADIR/`@HOSTNAME@`.pid
+  pid_file="$DATADIR/`@HOSTNAME@`.pid"
 else
   case "$pid_file" in
     /* ) ;;
@@ -491,9 +541,9 @@ fi
 #
 # If there exists an old pid file, check if the daemon is already running
 # Note: The switches to 'ps' may depend on your operating system
-if test -f $pid_file
+if test -f "$pid_file"
 then
-  PID=`cat $pid_file`
+  PID=`cat "$pid_file"`
   if @CHECK_PID@
   then
     if @FIND_PROC@
@@ -502,8 +552,8 @@ then
       exit 1
     fi
   fi
-  rm -f $pid_file
-  if test -f $pid_file
+  rm -f "$pid_file"
+  if test -f "$pid_file"
   then
     log_error "Fatal error: Can't remove the pid file:
 $pid_file
@@ -545,11 +595,17 @@ test -n "$NOHUP_NICENESS" && cmd="$cmd < /dev/null"
 log_notice "Starting $MYSQLD daemon with databases from $DATADIR"
 while true
 do
-  rm -f $safe_mysql_unix_port $pid_file	# Some extra safety
+  rm -f $safe_mysql_unix_port "$pid_file"	# Some extra safety
 
   eval_log_error "$cmd"
 
-  if test ! -f $pid_file		# This is removed if normal shutdown
+  if [ $want_syslog -eq 0 -a ! -e "$err_log" ]; then
+    touch "$err_log"                    # hypothetical: log was renamed but not
+    chown $user "$err_log"              # flushed yet. we'd recreate it with
+    chmod "$fmode" "$err_log"           # wrong owner next time we log, so set
+  fi                                    # it up correctly while we can!
+
+  if test ! -f "$pid_file"		# This is removed if normal shutdown
   then
     break
   fi
